# 
# -------------------------------------------------------------------
#                                   X-BONE
#
#                          http://www.isi.edu/xbone
#                USC Information Sciences Institute (USC/ISI)
#                   Marina del Rey, California 90292, USA
#                          Copyright (c) 1998-2005
# 
# -------------------------------------------------------------------
#
# Copyright (c) 1998-2005 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and
# its documentation in source and binary forms for non-commercial
# purposes and without fee is hereby granted, provided that the above
# copyright notice appear in all copies and that both the copyright
# notice and this permission notice appear in supporting
# documentation, and that any documentation, advertising materials,
# and other materials related to such distribution and use acknowledge
# that the software was developed by the University of Southern
# California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived
# from this software without specific prior written permission.
# 
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT
# THE SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS
# PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so
# noted when applicable.
#
# -------------------------------------------------------------------
#
# Effort partly sponsored by the Defense Advanced Research Projects
# Agency (DARPA) and Air Force Research Laboratory, Air Force Materiel
# Command, USAF, under agreement numbers F30602-98-1-0200 (X-Bone) and
# F30602-01-2-0529 (DynaBone). The views and conclusions contained
# herein are those of the authors and should not be interpreted as
# necessarily representing the official policies or endorsements,
# either expressed or implied, of the Defense Advanced Research
# Projects Agency (DARPA), the Air Force Research Laboratory, or the
# U.S. Government.
#
# This work was partly supported by the NSF STI-XTEND (ANI-0230789)
# and NETFS (ANI-0129689) projects. Any opinions, findings, and
# conclusions or recommendations expressed in this material are those
# of the authors and do not necessarily reflect the views of the
# National Science Foundation.
#
# -------------------------------------------------------------------
#  $RCSfile: Divert.pm,v $
#
# $Revision: 1.4 $
#   $Author: pingali $
#     $Date: 2005/03/31 07:03:43 $
#    $State: Exp $
# ----------------------------------------------------------------------------

# Primary Author: Lars Eggert

package Divert;

require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw();
@EXPORT_OK = qw(create destroy);

use strict;
use Utilities qw(:DEFAULT $verbose $debug);
use BARP;
use POSIX;
use IO::Socket::INET;


# Constants
#
use constant IPFW      => "/sbin/ipfw -q ";  # path to firewall comand


# Purpose:
#       Set up divert rules to capture all packets arriving over the
#       innerlays.
#       Must use sysdie/syswarn; this runs after we became a daemon.
#
# Arguments:
#       $outerlay_addr    outerlay IP address (Net::IP object)
#
#       $outerlay_block   outerlay block (Net::IP object)
#
# Returns:
#       ($fh, @rules)  a pair consisting of a non-blocking, read/write file
#                      handle to the divert socket, and an array of
#                      corresponding ipfw rule numbers used for divert rules
#
# Exceptions:
#       thrown on error
#
sub create ($$) {
  my ($outerlay_addr, $outerlay_block) = @_;
  my ($port, $fh);

  do {
    # Try random ports until we find an available one. Since the divert
    # port space is usually extremely sparse, this should work fine.
    $port = 1024 + int rand (65535 - 1024);
    
    # Open divert socket.
    sysprint "IO::Socket::INET->new($port, divert, SOCK_RAW)" if $verbose;
    $fh = IO::Socket::INET->new(LocalPort => $port,
				Proto     => "divert",
				Type      => SOCK_RAW);
    defined $fh 
      or syswarn "Cannot open divert socket on port $port, retrying: $!";

  } unless (defined $fh);

  # Try to find the two lowest-numbered, available ipfw rule numbers.
  my @rules;
  my %used;
  foreach my $line (IPFW ." show") {
    if ($line =~ /^(\d+)/) { $used{$1} = 1; }
  }
  my $needed = 2;
  for (my $rule = 1; $rule < 65535; $rule++) {
    unless (exists $used{$rule}) {
      push @rules, $rule;
      last if --$needed == 0;
    }
  }

  # We can already create the return value here, so we can mess with
  # the @rules list below.
  my @result = ($fh, @rules);

  # Now that we have an available port and rule numbers, add divert rules.
  my $rule = shift @rules;
  my %innerlay_ips = BARP::innerlay_ips $outerlay_addr;
  foreach my $innerlay (values %innerlay_ips) {
    my $cmd = IPFW. "add $rule divert $port ipencap from any to ". 
      $innerlay->ip ." in";
    sysprint $cmd if $verbose;
    system ($cmd) == 0 or sysdie "Cannot execute $cmd: $!";
  }
  
  # Make it non-blocking, otherwise later read calls will block.
  $fh->autoflush(1);
  fcntl $fh, F_SETFL, O_NONBLOCK
    or sysdie "Cannot fcntl divert, F_SEFTL, O_NONBLOCK: $!";

  # Punch a hole into the firewall, to allow all packets over the tun
  # to go through. This needs to occur *after* the divert rules are
  # added, so the decapsed packets pass by this rule and are let
  # through.
  $rule = shift @rules;
  my $block = $outerlay_block->ip . "/" . $outerlay_block->prefixlen;
  my $cmd = IPFW. "add $rule allow ip from $block to $block";
  sysprint $cmd if $verbose;
  system ($cmd) == 0 or sysdie "Cannot execute $cmd: $!";

  # Return the divert socket.
  return @result;
}



# Purpose:
#       Destroy the divert rule (undo divert_create operation).
#       Must use sysdie/syswarn; this runs after we became a daemon.
#
# Arguments:
#       $fh     file handle of the divert socket (return value from
#               divert_create)
#
# Returns:
#       1
#
# Exceptions:
#       -
#
sub destroy ($@) {
  my ($fh, @rules) = @_;

  # close file handle if it's open still
  $fh->close if defined $fh;

  # remove divert rules
  foreach my $rule (@rules) {
    my $cmd = IPFW. "delete $rule";
    sysprint $cmd if $verbose;
    system ($cmd) == 0 or syswarn "Cannot execute $cmd: $!";
  }
  
  return 1;
}



1;
